/*
 * Copyright (c) 2001
 *	Broadcom Corporation.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Broadcom Corporation.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Broadcom Corporation ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Broadcom Corporation
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#ifndef BC_MM_H
#define BC_MM_H

#include "opt_bcm5701.h"

#define PCIX_TARGET_WORKAROUND 1
#define INCLUDE_TBI_SUPPORT 1

#ifdef BC_PRIVATE_JUMBOS
#define INCLUDE_5701_AX_FIX 1
#endif

#define BCM_NIC_SEND_BD 1

#define T3_JUMBO_RCV_RCB_ENTRY_COUNT 256
#define BC_JUMBO_DESC_CNT 80
#define BC_JUMBO_FRAMELEN 9014
#define BC_JUMBO_MTU 9000

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/sockio.h>
#include <sys/mbuf.h>
#include <sys/malloc.h>
#include <sys/kernel.h>
#include <sys/sysctl.h>
#include <sys/socket.h>
#include <sys/queue.h>

#include <net/if.h>
#include <net/if_arp.h>
#include <net/ethernet.h>
#include <net/if_dl.h>
#include <net/if_media.h>

#include <net/bpf.h>

#if NVLAN > 0
#include <net/if_types.h>
#include <net/if_vlan_var.h>
#endif

#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>

#include <vm/vm.h>              /* for vtophys */
#include <vm/pmap.h>            /* for vtophys */
#include <machine/clock.h>      /* for DELAY */
#include <machine/bus_memio.h>
#include <machine/bus.h>
#include <machine/resource.h>
#include <machine/lock.h>
#include <sys/bus.h>
#include <sys/rman.h>

#include <pci/pcireg.h>
#include <pci/pcivar.h>

#if BYTE_ORDER == BIG_ENDIAN
#define BIG_ENDIAN_HOST 1
#define MM_SWAP_LE32(x) bswap32(x)
#else
#define MM_SWAP_LE32(x) x
#endif

#define MM_MEMWRITEL(sc, offset, val)				\
	bus_space_write_4(((struct bc_softc *) sc)->bc_btag,	\
	((struct bc_softc *) sc)->bc_bhandle, (bus_size_t) offset, val)

#define MM_MEMREADL(sc, offset)					\
	bus_space_read_4(((struct bc_softc *) sc)->bc_btag,	\
        ((struct bc_softc *) sc)->bc_bhandle, (bus_size_t) offset)

typedef int MM_ATOMIC_T;

#define MM_ATOMIC_SET(ptr, val) *(ptr) = (val)
#define MM_ATOMIC_READ(ptr) *(ptr)
#define MM_ATOMIC_INC(ptr) atomic_add_int(ptr, 1)
#define MM_ATOMIC_ADD(ptr, val) atomic_add_int(ptr, val)
#define MM_ATOMIC_DEC(ptr) atomic_subtract_int(ptr, 1)
#define MM_ATOMIC_SUB(ptr, val) atomic_subtract_int(ptr, val);

#define MM_MB() 
#define MM_WMB() 

#include "bc_lm.h"
#include "bc_queue.h"
#include "bc_tigon3.h"

extern int MM_Packet_Desc_Size;

#define ETHER_ALIGN	2
#define MM_PACKET_DESC_SIZE MM_Packet_Desc_Size

DECLARE_QUEUE_TYPE(UM_RX_PACKET_Q, MAX_RX_PACKET_DESC_COUNT+1);
#ifdef BC_PRIVATE_JUMBOS
#define BC_JUMBO_MEM_BLK (BC_JUMBO_DESC_CNT * 2)
DECLARE_QUEUE_TYPE(JUMBO_MEM_Q, BC_JUMBO_MEM_BLK+1);
#endif

#define MAX_MEM 16

#define BCM_VENDORID		0x14e4
#define BCM5700_DEVICEID	0x1644
#define BCM5701_DEVICEID	0x1645
#define BCM5702_DEVICEID	0x16a6
#define BCM5702x_DEVICEID	0x16c6
#define BCM5703_DEVICEID	0x16a7
#define BCM5703x_DEVICEID	0x16c7
#define BCM5704_DEVICEID	0x1648
#define BCM5704s_DEVICEID	0x16a8
#define BCM5705_DEVICEID	0x1653
#define BCM5705m_DEVICEID	0x165d

struct bc_type {
	u_int16_t		bc_vid;
	u_int16_t		bc_did;
	u_int16_t		bc_svid;
	u_int16_t		bc_ssid;
	char			*bc_name;
};


struct bc_softc {
	LM_DEVICE_BLOCK lm_dev;
	struct arpcom		arpcom;		/* interface info */
	bus_space_handle_t	bc_bhandle;
	vm_offset_t		bc_vhandle;
	bus_space_tag_t		bc_btag;
	void			*bc_intrhand;
	struct resource		*bc_irq;
	struct resource		*bc_res;
	struct ifmedia		ifmedia;	/* media info */
	u_int8_t		bc_unit;	/* interface number */
	device_t		dev;
	char *name;
	void *mem_list[MAX_MEM];
	int mem_size_list[MAX_MEM];
	int mem_list_num;
#ifdef BC_PRIVATE_JUMBOS
	void *jumbo_mem_ptr;
	int jumbo_mem_size;
#endif
	int index;
	int opened;
	int delayed_link_ind; /* Delay link status during initial load */
	int adapter_just_inited; /* the first few seconds after init. */
	int timer_interval;
	int adaptive_expiry;
	int tx_full;
	int tx_queued;
	int line_speed;		/* in Mbps, 0 if link is down */
	UM_RX_PACKET_Q rx_out_of_buf_q;
#ifdef BC_PRIVATE_JUMBOS
	JUMBO_MEM_Q jumbo_free_mem_q;
#endif
	int rx_out_of_buf;
	int rx_buf_repl_thresh;
	int rx_buf_repl_panic_thresh;
	int do_global_lock;
	struct simplelock global_lock;
	struct simplelock undi_lock;
	int undi_flags;
	volatile int interrupt;
	int rx_adaptive_coalesce;
	uint rx_last_cnt;
	uint tx_last_cnt;
	uint rx_curr_coalesce_frames;
	uint rx_curr_coalesce_ticks;
};

#ifdef BC_PRIVATE_JUMBOS
typedef struct bc_jmem {
	struct bc_softc *sc;
	int bc_inuse;
} bc_jmem_t;
#endif

#define MAX_FRAGS 64

typedef struct _UM_PACKET {
	LM_PACKET lm_packet;
	struct mbuf *mbuf;
	LM_FRAG_LIST frag_list;
        LM_FRAG frag_list_bufs[MAX_FRAGS];
} UM_PACKET, *PUM_PACKET;

#ifdef SMP
#define MM_ACQUIRE_UNDI_LOCK(_pDevice) \
	if (!(((struct bc_softc *)(_pDevice))->do_global_lock)) {	\
		int s;						\
		s = splimp();					\
		s_lock(&((struct bc_softc *)(_pDevice))->undi_lock);	\
		((struct bc_softc *)(_pDevice))->undi_flags = s; \
	}

#define MM_RELEASE_UNDI_LOCK(_pDevice) \
	if (!(((struct bc_softc *)(_pDevice))->do_global_lock)) {	\
		int s = ((struct bc_softc *) (_pDevice))->undi_flags; \
		s_unlock(&((struct bc_softc *)(_pDevice))->undi_lock); \
		splx(s);					\
	}
#else
#define MM_ACQUIRE_UNDI_LOCK(_pDevice) \
	if (!(((struct bc_softc *)(_pDevice))->do_global_lock)) {	\
		((struct bc_softc *)(_pDevice))->undi_flags = splimp(); \
	}

#define MM_RELEASE_UNDI_LOCK(_pDevice) \
	if (!(((struct bc_softc *)(_pDevice))->do_global_lock)) {	\
		splx(((struct bc_softc *) (_pDevice))->undi_flags); \
	}
#endif

#define MM_ACQUIRE_PHY_LOCK_IN_IRQ(_pDevice)
#define MM_RELEASE_PHY_LOCK_IN_IRQ(_pDevice)

#define MM_UINT_PTR(_ptr)   ((unsigned long) (_ptr))

#define DbgPrint(fmt, arg...) printf(fmt, ##arg)
#define DbgBreakPoint()
#define MM_Wait(time) DELAY(time)

#endif /* BC_MM_H */
