/*
 * Copyright (c) 2001
 *	Broadcom Corporation.  All rights reserved.
 *
 * Portions of this file were taken from if_ti.c written by Bill Paul.
 * Copyright (c) 1997, 1998, 1999
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Broadcom Corporation.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Broadcom Corporation ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Broadcom Corporation
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $FreeBSD: src/sys/pci/if_bc.c,v 2.0.2 2003/06/19 00:00:00 bcm Exp $
 */

/*
 * Broadcom Corporation BCM570X PCI-X gigabit ethernet driver for FreeBSD.
 * 
 * Written by Michael Chan <mchan@broadcom.com>
 */

#include "bc_mm.h"

#include <sys/eventhandler.h>
#include <sys/reboot.h>

#define DEBUG 1

#ifdef IP_GDB
#include <ipgdb/ipgdb.h>
#include <sys/reboot.h>	/* for RB_GDB declaration */
#endif

#define BC_CSUM_FEATURES	(CSUM_IP | CSUM_TCP | CSUM_UDP)
/* Version 2.0.1 updated driver */
/* Version 2.0.0 supports 5704/5705 as well */
/* Version 1.0.0 supports 5703/5702 */
#define BC_VERSION_STRING	"2.0.1"

#if !defined(lint)
static const char rcsid[] =
  "$FreeBSD: src/sys/pci/if_bc.c,v 2.0.1 2003/05/02 01:00:00 bcm Exp $";
#endif

/*
 * Various supported device vendors/types and their names.
 */

static struct bc_type bc_devs[] = {
	{ BCM_VENDORID, BCM5700_DEVICEID, BCM_VENDORID, BCM5700_DEVICEID,
		"Broadcom NetXtreme BCM5700 Gigabit Ethernet" },
	{ BCM_VENDORID, BCM5701_DEVICEID, BCM_VENDORID, 0x0001,
		"Broadcom NetXtreme BCM5701 Gigabit Ethernet" },
	{ BCM_VENDORID, BCM5701_DEVICEID, BCM_VENDORID, 0x0005,
		"Broadcom NetXtreme BCM5701 Gigabit Ethernet" },
	{ BCM_VENDORID, BCM5701_DEVICEID, BCM_VENDORID, 0x0006,
		"Broadcom NetXtreme BCM5701 Gigabit Ethernet" },
	{ BCM_VENDORID, BCM5701_DEVICEID, BCM_VENDORID, 0x0007,
		"Broadcom NetXtreme BCM5701 Gigabit Ethernet" },
	{ BCM_VENDORID, BCM5701_DEVICEID, BCM_VENDORID, 0x0008,
		"Broadcom NetXtreme BCM5701 Gigabit Ethernet" },
	{ BCM_VENDORID, BCM5702_DEVICEID, BCM_VENDORID, 0x8009,
		"Broadcom NetXtreme BCM5702 Gigabit Ethernet" },
	{ BCM_VENDORID, BCM5702_DEVICEID, BCM_VENDORID, 0x000c,
		"Broadcom NetXtreme BCM5702 Gigabit Ethernet" },
	{ BCM_VENDORID, BCM5702x_DEVICEID, BCM_VENDORID, 0x8009,
		"Broadcom NetXtreme BCM5702 Gigabit Ethernet" },
	{ BCM_VENDORID, BCM5702x_DEVICEID, BCM_VENDORID, 0x000c,
		"Broadcom NetXtreme BCM5702 Gigabit Ethernet" },
	{ BCM_VENDORID, BCM5703_DEVICEID, BCM_VENDORID, 0x0009,
		"Broadcom NetXtreme BCM5703 Gigabit Ethernet" },
	{ BCM_VENDORID, BCM5703_DEVICEID, BCM_VENDORID, 0x000a,
		"Broadcom NetXtreme BCM5703 Gigabit Ethernet" },
	{ BCM_VENDORID, BCM5703_DEVICEID, BCM_VENDORID, 0x000b,
		"Broadcom NetXtreme BCM5703 Gigabit Ethernet" },
	{ BCM_VENDORID, BCM5703x_DEVICEID, BCM_VENDORID, 0x0009,
		"Broadcom NetXtreme BCM5703 Gigabit Ethernet" },
	{ BCM_VENDORID, BCM5703x_DEVICEID, BCM_VENDORID, 0x000a,
		"Broadcom NetXtreme BCM5703 Gigabit Ethernet" },
	{ BCM_VENDORID, BCM5703x_DEVICEID, BCM_VENDORID, 0x000b,
		"Broadcom NetXtreme BCM5703 Gigabit Ethernet" },
	{ BCM_VENDORID, BCM5704_DEVICEID, BCM_VENDORID, BCM5704_DEVICEID,
		"Broadcom NetXtreme BCM5704 Gigabit Ethernet" },
	{ BCM_VENDORID, BCM5704s_DEVICEID, BCM_VENDORID, BCM5704s_DEVICEID,
		"Broadcom NetXtreme BCM5704s Gigabit Ethernet" },
	{ BCM_VENDORID, BCM5705_DEVICEID, BCM_VENDORID, BCM5705_DEVICEID,
		"Broadcom NetXtreme BCM5705 Gigabit Ethernet" },
	{ BCM_VENDORID, BCM5705m_DEVICEID, BCM_VENDORID, BCM5705m_DEVICEID,
		"Broadcom NetXtreme BCM5705m Gigabit Ethernet" },
	{ 0, 0, 0, 0, NULL }
};


static int bc_probe		__P((device_t));
static int bc_attach		__P((device_t));
static int bc_detach		__P((device_t));

#ifdef IP_GDB_BC
void bc_intr		__P((void *));
#else
static void bc_intr		__P((void *));
#endif
static void bc_start		__P((struct ifnet *));
static int bc_ioctl		__P((struct ifnet *, u_long, caddr_t));
static void bc_init		__P((void *));
static void bc_watchdog		__P((struct ifnet *));
static void bc_shutdown		__P((device_t));
static int bc_ifmedia_upd	__P((struct ifnet *));
static void bc_ifmedia_sts	__P((struct ifnet *, struct ifmediareq *));

static void bc_setmulti		__P((struct bc_softc *));

static int bc_newbuf		__P((struct bc_softc *, PUM_PACKET));
static int bc_freemem		__P((struct bc_softc *));

#ifdef BC_PRIVATE_JUMBOS
static int bc_alloc_jumbo_mem	__P((struct bc_softc *));
static int bc_free_jumbo_mem	__P((struct bc_softc *));
static void *bc_jalloc		__P((struct bc_softc *));
static void bc_jfree		__P((caddr_t, u_int));
static void bc_jref		__P((caddr_t, u_int));
#endif /* BC_PRIVATE_JUMBOS */

#ifdef DEBUG
struct ifrdebug {
	char	ifr_name[IFNAMSIZ];
	unsigned long addr;
	unsigned long value;
};
#define SIOCGIFREADMEM _IOWR('i', 100, struct ifrdebug)
#endif
struct if_bc_stat {
	char    ifr_name[IFNAMSIZ];
	T3_STATS_BLOCK	nic_stats;
};
#define SIOCGIFGETSTATS _IOWR('i', 101, struct if_bc_stat)



static device_method_t bc_methods[] = {
	/* Device interface */
	DEVMETHOD(device_probe,		bc_probe),
	DEVMETHOD(device_attach,	bc_attach),
	DEVMETHOD(device_detach,	bc_detach),
	DEVMETHOD(device_shutdown,	bc_shutdown),
	{ 0, 0 }
};

static driver_t bc_driver = {
	"bc",
	bc_methods,
	sizeof(struct bc_softc)
};

static devclass_t bc_devclass;

int MM_Packet_Desc_Size = sizeof(UM_PACKET);

DRIVER_MODULE(if_bc, pci, bc_driver, bc_devclass, 0, 0);

/*
 * Configure the  multicast address filter.
 */
static void bc_setmulti(sc)
	struct bc_softc		*sc;
{
	PLM_DEVICE_BLOCK	pDevice = (PLM_DEVICE_BLOCK) sc;
	struct ifnet		*ifp;
	struct ifmultiaddr	*ifma;

	ifp = &sc->arpcom.ac_if;

	if (ifp->if_flags & IFF_ALLMULTI) {
		LM_SetReceiveMask(pDevice,
			pDevice->ReceiveMask | LM_ACCEPT_ALL_MULTICAST);
		return;
	} else {
		LM_SetReceiveMask(pDevice,
			pDevice->ReceiveMask & ~LM_ACCEPT_ALL_MULTICAST);
	}


	LM_MulticastClear(pDevice);
	/* Now program new ones. */
	for (ifma = ifp->if_multiaddrs.lh_first;
	    ifma != NULL; ifma = ifma->ifma_link.le_next) {
		if (ifma->ifma_addr->sa_family != AF_LINK)
			continue;
		LM_MulticastAdd(pDevice, 
			LLADDR((struct sockaddr_dl *)ifma->ifma_addr));
	}

	return;
}


/*
 * Probe for a Broadcom chip. Check the PCI vendor and device IDs
 * against our list and return its name if we find a match.
 */
static int bc_probe(dev)
	device_t		dev;
{
	struct bc_type		*b;

	b = bc_devs;

	while(b->bc_name != NULL) {
		if ((pci_get_vendor(dev) == b->bc_vid) &&
			(pci_get_device(dev) == b->bc_did) &&
			(pci_read_config(dev, PCIR_SUBVEND_0, 2) ==
				b->bc_svid) &&
			(pci_read_config(dev, PCIR_SUBDEV_0, 2) ==
				b->bc_ssid)) {

			device_set_desc(dev, b->bc_name);
			return(0);
		}
		b++;
	}

	return(ENXIO);
}

static int bc_attach(dev)
	device_t		dev;
{
	int			s;
	struct ifnet		*ifp;
	struct bc_softc		*sc;
	PLM_DEVICE_BLOCK	pDevice;
	int			unit, error = 0, rid;

	s = splimp();

	sc = device_get_softc(dev);
	unit = device_get_unit(dev);
	bzero(sc, sizeof(struct bc_softc));
	pDevice = &sc->lm_dev;

	sc->bc_unit = unit;
	sc->dev = dev;
	if (LM_GetAdapterInfo(pDevice) != LM_STATUS_SUCCESS) {
		printf("bc%d: Get Adapter info failed!\n", sc->bc_unit);
		error = ENXIO;
		goto fail;
	}

	/* Allocate interrupt */
	rid = 0;
	
	sc->bc_irq = bus_alloc_resource(dev, SYS_RES_IRQ, &rid, 0, ~0, 1,
	    RF_SHAREABLE | RF_ACTIVE);

	if (sc->bc_irq == NULL) {
		printf("bc%d: couldn't map interrupt\n", unit);
		error = ENXIO;
		goto fail;
	}

	error = bus_setup_intr(dev, sc->bc_irq, INTR_TYPE_NET,
	   bc_intr, sc, &sc->bc_intrhand);

	if (error) {
		bus_release_resource(dev, SYS_RES_IRQ, 0, sc->bc_irq);
		printf("bc%d: couldn't set up irq\n", unit);
		goto fail;
	}
#ifdef BC_PRIVATE_JUMBOS
	if (bc_alloc_jumbo_mem(sc)) {
		bus_teardown_intr(dev, sc->bc_irq, sc->bc_intrhand);
		bus_release_resource(dev, SYS_RES_IRQ, 0, sc->bc_irq);
		printf("bc%d: jumbo buffer allocation failed\n", sc->bc_unit);
		error = ENOMEM;
		goto fail;
	}
#endif /* BC_PRIVATE_JUMBOS */
	QQ_InitQueue(&sc->rx_out_of_buf_q.Container, MAX_RX_PACKET_DESC_COUNT);

	if (LM_InitializeAdapter(pDevice) != LM_STATUS_SUCCESS) {
		bus_teardown_intr(dev, sc->bc_irq, sc->bc_intrhand);
		bus_release_resource(dev, SYS_RES_IRQ, 0, sc->bc_irq);
		bus_release_resource(dev, SYS_RES_MEMORY, PCIR_MAPS, sc->bc_res);
		bc_freemem(sc);
#ifdef BC_PRIVATE_JUMBOS
		bc_free_jumbo_mem(sc);
#endif
		printf("bc%d: Initialize Adapter failed!\n", sc->bc_unit);
		error = ENXIO;
		goto fail;
	}

#ifdef DEBUG
	printf("bc%d: sc address = %x\n", sc->bc_unit, (unsigned int) sc);
#endif
	bcopy(pDevice->NodeAddress, (caddr_t)&sc->arpcom.ac_enaddr, 6);

	/*
	 * A Broadcom chip was detected. Inform the world.
	 */
	printf("bc%d: Ethernet address: %6D Version %s\n", unit,
				sc->arpcom.ac_enaddr, ":", 
				BC_VERSION_STRING);

	/* Set up ifnet structure */
	ifp = &sc->arpcom.ac_if;
	ifp->if_softc = sc;
	ifp->if_unit = sc->bc_unit;
	ifp->if_name = "bc";
	ifp->if_flags = IFF_BROADCAST | IFF_SIMPLEX | IFF_MULTICAST;
	ifp->if_ioctl = bc_ioctl;
	ifp->if_output = ether_output;
	ifp->if_start = bc_start;
	ifp->if_watchdog = bc_watchdog;
	ifp->if_init = bc_init;
	ifp->if_mtu = ETHERMTU;
	ifp->if_snd.ifq_maxlen = pDevice->TxPacketDescCnt - 1;
	ifp->if_hwassist = BC_CSUM_FEATURES;

	/* Set up ifmedia support. */
	ifmedia_init(&sc->ifmedia, IFM_IMASK, bc_ifmedia_upd, bc_ifmedia_sts);
	if (!pDevice->EnableTbi) {
		/*
		 * Copper cards allow manual 10/100 mode selection,
		 * but not manual 1000baseTX mode selection. Why?
		 * Becuase currently there's no way to specify the
		 * master/slave setting through the firmware interface,
		 * so Alteon decided to just bag it and handle it
		 * via autonegotiation.
		 */
		ifmedia_add(&sc->ifmedia, IFM_ETHER|IFM_10_T, 0, NULL);
		ifmedia_add(&sc->ifmedia,
		    IFM_ETHER|IFM_10_T|IFM_FDX, 0, NULL);
		ifmedia_add(&sc->ifmedia, IFM_ETHER|IFM_100_TX, 0, NULL);
		ifmedia_add(&sc->ifmedia,
		    IFM_ETHER|IFM_100_TX|IFM_FDX, 0, NULL);
		ifmedia_add(&sc->ifmedia, IFM_ETHER|IFM_1000_TX, 0, NULL);
		ifmedia_add(&sc->ifmedia,
		    IFM_ETHER|IFM_1000_TX|IFM_FDX, 0, NULL);
	} else {
		/* Fiber cards don't support 10/100 modes. */
		ifmedia_add(&sc->ifmedia, IFM_ETHER|IFM_1000_SX, 0, NULL);
		ifmedia_add(&sc->ifmedia,
		    IFM_ETHER|IFM_1000_SX|IFM_FDX, 0, NULL);
	}
	ifmedia_add(&sc->ifmedia, IFM_ETHER|IFM_AUTO, 0, NULL);
	ifmedia_set(&sc->ifmedia, IFM_ETHER|IFM_AUTO);

	/*
	 * Call MI attach routine.
	 */
	ether_ifattach(ifp, ETHER_BPF_SUPPORTED);
	LM_EnableInterrupt(pDevice);

fail:
	splx(s);

	return(error);
}

static int bc_detach(dev)
	device_t		dev;
{
	struct bc_softc		*sc;
	PLM_DEVICE_BLOCK	pDevice;
	struct ifnet		*ifp;
	int			s;

	s = splimp();

	sc = device_get_softc(dev);
	pDevice = &sc->lm_dev;
	ifp = &sc->arpcom.ac_if;

	ether_ifdetach(ifp, ETHER_BPF_SUPPORTED);
	LM_Halt(pDevice);
	pDevice->InitDone = FALSE;

	bus_teardown_intr(dev, sc->bc_irq, sc->bc_intrhand);
	bus_release_resource(dev, SYS_RES_IRQ, 0, sc->bc_irq);
	bus_release_resource(dev, SYS_RES_MEMORY, PCIR_MAPS, sc->bc_res);
	bc_freemem(sc);
#ifdef BC_PRIVATE_JUMBOS
	bc_free_jumbo_mem(sc);
#endif

	ifmedia_removeall(&sc->ifmedia);

	splx(s);

	return(0);
}

#ifdef IP_GDB_BC
void bc_intr(xsc)
#else
static void bc_intr(xsc)
#endif
	void			*xsc;
{
	struct bc_softc		*sc = xsc;
	PLM_DEVICE_BLOCK	pDevice = &sc->lm_dev;
	struct ifnet		*ifp;
	LM_UINT32 oldtag, newtag;

	ifp = &sc->arpcom.ac_if;

	if (pDevice->UseTaggedStatus) {
		if ((pDevice->pStatusBlkVirt->Status & STATUS_BLOCK_UPDATED) ||
			sc->adapter_just_inited) {
			MB_REG_WR(pDevice, Mailbox.Interrupt[0].Low, 1);
			oldtag = pDevice->pStatusBlkVirt->StatusTag;

			while (1) {
   				pDevice->pStatusBlkVirt->Status &=
					~STATUS_BLOCK_UPDATED;

				LM_ServiceInterrupts(pDevice);
				newtag = pDevice->pStatusBlkVirt->StatusTag;
				if (newtag == oldtag) {
					MB_REG_WR(pDevice,
						Mailbox.Interrupt[0].Low,
						oldtag << 24);
					pDevice->LastTag = oldtag;
					break;
				}
				oldtag = newtag;
			}
		}
	}
	else if ((pDevice->pStatusBlkVirt->Status & STATUS_BLOCK_UPDATED) ||
			sc->adapter_just_inited) {

		do {
			uint dummy;

			MB_REG_WR(pDevice, Mailbox.Interrupt[0].Low, 1);
   			pDevice->pStatusBlkVirt->Status &= ~STATUS_BLOCK_UPDATED;
			LM_ServiceInterrupts(pDevice);
			MB_REG_WR(pDevice, Mailbox.Interrupt[0].Low, 0);
			dummy = MB_REG_RD(pDevice, Mailbox.Interrupt[0].Low);
		}
		while (pDevice->pStatusBlkVirt->Status & STATUS_BLOCK_UPDATED);
	}

 	if ((ifp->if_snd.ifq_head != NULL) &&
		((ifp->if_flags & (IFF_RUNNING | IFF_OACTIVE)) == IFF_RUNNING))
	{
		bc_start(ifp);
	}
#ifdef IP_GDB_BC
	if (ipgdb_connected == 2) {
		ipgdb_connected = 1;
		boothowto |= RB_GDB; /* causes Debugger() to go remote */
		Debugger("remote debugging initiated");
	}
#endif
	return;
}

/*
 * Main transmit routine. To avoid having to do mbuf copies, we put pointers
 * to the mbuf data regions directly in the transmit descriptors.
 */
static void bc_start(ifp)
	struct ifnet		*ifp;
{
	struct bc_softc		*sc = ifp->if_softc;
	PLM_DEVICE_BLOCK	pDevice = &sc->lm_dev;
	struct mbuf		*m_head, *m;
	PLM_PACKET		pPacket;
	PUM_PACKET		pUmPacket;
	PLM_FRAG_LIST		pfrag_list;
	int			i;
	int			total_len;

	if (pDevice->LinkStatus == LM_STATUS_LINK_DOWN) {
		return;
	}

	IF_DEQUEUE(&ifp->if_snd, m_head);
	while (m_head) {

		pPacket = (PLM_PACKET)
			QQ_PopHead(&pDevice->TxPacketFreeQ.Container);
		if (pPacket == 0) {
			IF_PREPEND(&ifp->if_snd, m_head);
			ifp->if_flags |= IFF_OACTIVE;
			break;
		}
		pUmPacket = (PUM_PACKET) pPacket;
		pUmPacket->mbuf = m_head;
		pfrag_list = &pUmPacket->frag_list;
		pPacket->u.Tx.pFraglist = pfrag_list;
		pPacket->Flags = 0;
		if (m_head->m_pkthdr.csum_flags) {
			if (m_head->m_pkthdr.csum_flags & CSUM_IP)
				pPacket->Flags |= SND_BD_FLAG_IP_CKSUM;
			if (m_head->m_pkthdr.csum_flags & (CSUM_TCP | CSUM_UDP))
				pPacket->Flags |= SND_BD_FLAG_TCP_UDP_CKSUM;
		}

		total_len = 0;
		for (m = m_head, i = 0; m != NULL; m = m->m_next) {
			if (m->m_len == 0)
				continue;
if (i > 64) {
printf("bc%d: too many frags\n", sc->bc_unit);
break;
}
			pfrag_list->Fragments[i].FragSize = m->m_len;
			total_len += m->m_len;
			pfrag_list->Fragments[i].FragBuf.Low =
				vtophys(mtod(m, vm_offset_t));
			i++;
		}
		if (pDevice->SendBdLeft < i) {
			QQ_PushHead(&pDevice->TxPacketFreeQ.Container, pPacket);
			IF_PREPEND(&ifp->if_snd, m_head);
			ifp->if_flags |= IFF_OACTIVE;
			break;
		}
		pPacket->u.Tx.FragCount = i;
#if NVLAN > 0
		if ((m_head->m_flags & (M_PROTO1|M_PKTHDR)) ==
			(M_PROTO1|M_PKTHDR) &&
			m_head->m_pkthdr.rcvif != NULL &&
			m_head->m_pkthdr.rcvif->if_type == IFT_8021_VLAN) {

			struct ifvlan		*ifv;

			if ((ifv = m_head->m_pkthdr.rcvif->if_softc) != NULL) {
                		pPacket->Flags |= SND_BD_FLAG_VLAN_TAG;
				pPacket->VlanTag = ifv->ifv_tag;
			}
		}
#endif
		LM_SendPacket(pDevice, pPacket);

		/*
		 * If there's a BPF listener, bounce a copy of this frame
		 * to him.
		 */
		if (ifp->if_bpf)
			bpf_mtap(ifp, m_head);

		IF_DEQUEUE(&ifp->if_snd, m_head);
	}

	/*
	 * Set a timeout in case the chip goes out to lunch.
	 */
	ifp->if_timer = 5;

	return;
}

static void bc_init(xsc)
	void			*xsc;
{
	struct bc_softc		*sc = xsc;
	PLM_DEVICE_BLOCK	pDevice = &sc->lm_dev;
        int			s;
	struct ifnet		*ifp = &sc->arpcom.ac_if;

	s = splimp();

	LM_ResetAdapter(pDevice);
	LM_EnableInterrupt(pDevice);

	ifp->if_flags |= IFF_RUNNING;
	ifp->if_flags &= ~IFF_OACTIVE;

	splx(s);

	return;
}

/*
 * Set media options.
 */
static int bc_ifmedia_upd(ifp)
	struct ifnet		*ifp;
{
	struct bc_softc		*sc = ifp->if_softc;
	struct ifmedia		*ifm = &sc->ifmedia;
	PLM_DEVICE_BLOCK	pDevice = (PLM_DEVICE_BLOCK) sc;

	if (IFM_TYPE(ifm->ifm_media) != IFM_ETHER)
		return(EINVAL);

	switch(IFM_SUBTYPE(ifm->ifm_media)) {
	case IFM_AUTO:
		pDevice->RequestedLineSpeed = LM_LINE_SPEED_AUTO;
		pDevice->RequestedDuplexMode = LM_DUPLEX_MODE_UNKNOWN;
		pDevice->DisableAutoNeg = FALSE;
		break;
	case IFM_1000_SX:
	case IFM_1000_TX:
		pDevice->RequestedLineSpeed = LM_LINE_SPEED_1000MBPS;
		if ((ifm->ifm_media & IFM_GMASK) == IFM_FDX) {
			pDevice->RequestedDuplexMode = LM_DUPLEX_MODE_FULL;
		}
		else {
			pDevice->RequestedDuplexMode = LM_DUPLEX_MODE_HALF;
		}
		pDevice->DisableAutoNeg = TRUE;
		break;
	case IFM_100_FX:
	case IFM_100_TX:
		pDevice->RequestedLineSpeed = LM_LINE_SPEED_100MBPS;
		if ((ifm->ifm_media & IFM_GMASK) == IFM_FDX) {
			pDevice->RequestedDuplexMode = LM_DUPLEX_MODE_FULL;
		}
		else {
			pDevice->RequestedDuplexMode = LM_DUPLEX_MODE_HALF;
		}
		pDevice->DisableAutoNeg = TRUE;
		break;
	case IFM_10_FL:
	case IFM_10_T:
		pDevice->RequestedLineSpeed = LM_LINE_SPEED_100MBPS;
		if ((ifm->ifm_media & IFM_GMASK) == IFM_FDX) {
			pDevice->RequestedDuplexMode = LM_DUPLEX_MODE_FULL;
		}
		else {
			pDevice->RequestedDuplexMode = LM_DUPLEX_MODE_HALF;
		}
		pDevice->DisableAutoNeg = TRUE;
		break;
	}

	LM_SetupPhy(pDevice);
	return(0);
}

/*
 * Report current media status.
 */
static void bc_ifmedia_sts(ifp, ifmr)
	struct ifnet		*ifp;
	struct ifmediareq	*ifmr;
{
	struct bc_softc		*sc = ifp->if_softc;
	PLM_DEVICE_BLOCK	pDevice = (PLM_DEVICE_BLOCK) sc;

	ifmr->ifm_status = IFM_AVALID;
	ifmr->ifm_active = IFM_ETHER;

	if (pDevice->LinkStatus == LM_STATUS_LINK_DOWN)
		return;

	ifmr->ifm_status |= IFM_ACTIVE;

	if (pDevice->LineSpeed == LM_LINE_SPEED_1000MBPS) {
		if (pDevice->EnableTbi) {
			ifmr->ifm_active |= IFM_1000_SX;
		}
		else {
			ifmr->ifm_active |= IFM_1000_TX;
		}
	}
	else if (pDevice->LineSpeed == LM_LINE_SPEED_100MBPS) {
		ifmr->ifm_active |= IFM_100_TX;
	}
	else if (pDevice->LineSpeed == LM_LINE_SPEED_10MBPS) {
		ifmr->ifm_active |= IFM_10_T;
	}
	if (pDevice->DuplexMode == LM_DUPLEX_MODE_FULL) {
		ifmr->ifm_active |= IFM_FDX;
	}
	else {
		ifmr->ifm_active |= IFM_HDX;
	}
	
	return;
}

static int bc_ioctl(ifp, command, data)
	struct ifnet		*ifp;
	u_long			command;
	caddr_t			data;
{
	struct bc_softc		*sc = ifp->if_softc;
	PLM_DEVICE_BLOCK	pDevice = (PLM_DEVICE_BLOCK) sc;
	struct ifreq		*ifr = (struct ifreq *) data;
	int			s, error = 0;

	s = splimp();

	switch(command) {
	case SIOCSIFADDR:
	case SIOCGIFADDR:
		error = ether_ioctl(ifp, command, data);
		break;
	case SIOCSIFMTU:
		if (ifr->ifr_mtu > BC_JUMBO_MTU)
			error = EINVAL;
		else {
			ifp->if_mtu = ifr->ifr_mtu;
			LM_Halt(pDevice);
			pDevice->InitDone = FALSE;
			bc_freemem(sc);
			pDevice->RxMtu = ifp->if_mtu + 14;
			pDevice->TxMtu = pDevice->RxMtu;
			if (ifp->if_mtu > 1514) {
				pDevice->RxJumboDescCnt = BC_JUMBO_DESC_CNT;
				pDevice->RxPacketDescCnt = 
					pDevice->RxJumboDescCnt +
					pDevice->RxStdDescCnt;
        			pDevice->RxJumboBufferSize =
					(pDevice->RxMtu + 8 /* CRC + VLAN */ +
            				31) & ~32;
			}
			LM_InitializeAdapter(pDevice);
			LM_EnableInterrupt(pDevice);
		}
		break;
	case SIOCSIFFLAGS:
		if (ifp->if_flags & IFF_UP) {
			/*
			 * If only the state of the PROMISC flag changed,
			 * then just use the 'set promisc mode' command
			 * instead of reinitializing the entire NIC. Doing
			 * a full re-init means reloading the firmware and
			 * waiting for it to start up, which may take a
			 * second or two.
			 */

			if (ifp->if_flags & IFF_RUNNING &&
				ifp->if_flags & IFF_PROMISC &&
				!(pDevice->ReceiveMask & LM_PROMISCUOUS_MODE)) {

				LM_SetReceiveMask(pDevice, pDevice->ReceiveMask|
					 LM_PROMISCUOUS_MODE);
			} else if (ifp->if_flags & IFF_RUNNING &&
				!(ifp->if_flags & IFF_PROMISC) &&
				pDevice->ReceiveMask & LM_PROMISCUOUS_MODE) {

				LM_SetReceiveMask(pDevice, pDevice->ReceiveMask&
					 ~LM_PROMISCUOUS_MODE);
			} else
				bc_init(sc);
		} else {
			if (ifp->if_flags & IFF_RUNNING) {
				LM_Abort(pDevice);
			}
		}
		error = 0;
		break;
	case SIOCADDMULTI:
	case SIOCDELMULTI:
		if (ifp->if_flags & IFF_RUNNING) {
			bc_setmulti(sc);
			error = 0;
		}
		break;
	case SIOCSIFMEDIA:
	case SIOCGIFMEDIA:
		error = ifmedia_ioctl(ifp, ifr, &sc->ifmedia, command);
		break;
#ifdef DEBUG
	case SIOCGIFREADMEM: {
		struct ifrdebug *ifrd = (struct ifrdebug *) data;

		if (ifrd->addr > 0x38800)
			error = EINVAL;
		if (ifrd->addr >= 0x30000)
			ifrd->value = LM_RegRdInd(&sc->lm_dev, ifrd->addr);
		else
			ifrd->value = LM_MemRdInd(&sc->lm_dev, ifrd->addr);
		error = 0;
		break;
	}
#endif
	case SIOCGIFGETSTATS: {
		struct if_bc_stat *if_stat = (struct if_bc_stat *) data;

		if_stat->nic_stats = *(pDevice->pStatsBlkVirt);
		error = 0;
		break;
	}
	default:
		error = EINVAL;
		break;
	}

	(void)splx(s);

	return(error);
}

static void bc_watchdog(ifp)
	struct ifnet		*ifp;
{
	struct bc_softc		*sc = ifp->if_softc;
	PLM_DEVICE_BLOCK	pDevice = &sc->lm_dev;

	printf("bc%d: watchdog timeout -- resetting\n", sc->bc_unit);
	LM_ResetAdapter(pDevice);
	LM_EnableInterrupt(pDevice);

	ifp->if_oerrors++;

	return;
}

/*
 * Stop all chip I/O so that the kernel's probe routines don't
 * get confused by errant DMAs when rebooting.
 */
static void bc_shutdown(dev)
	device_t		dev;
{
	struct bc_softc		*sc = device_get_softc(dev);
	PLM_DEVICE_BLOCK	pDevice = &sc->lm_dev;

	LM_Halt(pDevice);
	pDevice->InitDone = FALSE;

	return;
}

/*
 * Intialize a standard receive ring descriptor.
 */
static int bc_newbuf(sc, pUmPacket)
	struct bc_softc		*sc;
	PUM_PACKET		pUmPacket;
{
	struct mbuf		*m_new = NULL;
	PLM_PACKET		pPacket = &pUmPacket->lm_packet;

	MGETHDR(m_new, M_DONTWAIT, MT_DATA);
	if (m_new == NULL) {
		printf("bc%d: mbuf allocation failed!\n", sc->bc_unit);
		return(ENOBUFS);
	}

	if (pPacket->u.Rx.RxBufferSize <= (MCLBYTES - ETHER_ALIGN)) {
		MCLGET(m_new, M_DONTWAIT);
		if (!(m_new->m_flags & M_EXT)) {
			printf("bc%d: cluster allocation failed "
				"-- packet dropped!\n", sc->bc_unit);
			m_freem(m_new);
			return(ENOBUFS);
		}
		m_new->m_len = m_new->m_pkthdr.len = MCLBYTES;
	}
	else {
#ifdef BC_PRIVATE_JUMBOS
		caddr_t			*buf = NULL;

		/* Allocate the jumbo buffer */
		buf = bc_jalloc(sc);
		if (buf == NULL) {
			m_freem(m_new);
			printf("bc%d: jumbo allocation failed "
				"-- packet dropped!\n", sc->bc_unit);
			return(ENOBUFS);
		}

		/* Attach the buffer to the mbuf. */
		m_new->m_data = m_new->m_ext.ext_buf = (void *)buf;
		m_new->m_flags |= M_EXT;
		m_new->m_len = m_new->m_pkthdr.len =
			m_new->m_ext.ext_size = BC_JUMBO_FRAMELEN;
		m_new->m_ext.ext_free = bc_jfree;
		m_new->m_ext.ext_ref = bc_jref;
#endif /* BC_PRIVATE_JUMBOS */
	}
	pUmPacket->mbuf = m_new;

	m_adj(m_new, ETHER_ALIGN);
	pPacket->u.Rx.pRxBufferVirt = mtod(m_new, caddr_t);
	pPacket->u.Rx.RxBufferPhy.High = 0;
	pPacket->u.Rx.RxBufferPhy.Low = vtophys(mtod(m_new, caddr_t));

	return(0);
}

static int
bc_freemem(struct bc_softc *sc)
{
	int i, size;

	for (i = 0; i < sc->mem_list_num; i++) {
		if ((size = sc->mem_size_list[i]) != 0) {
			contigfree(sc->mem_list[i], size, M_DEVBUF);
		}
		else {
			free(sc->mem_list[i], M_DEVBUF);
		}
	}
	sc->mem_list_num = 0;
	return 0;
}

#ifdef BC_PRIVATE_JUMBOS

static int bc_alloc_jumbo_mem(sc)
	struct bc_softc		*sc;
{
	caddr_t			ptr;
	register int		i;
	int			size;
	bc_jmem_t		*jmem_ptr;

	/* Grab a big chunk o' storage. */
	size = BC_JUMBO_MEM_BLK * (BC_JUMBO_FRAMELEN + ETHER_ALIGN +
		sizeof(bc_jmem_t));
	if (size == 0)
		return 0;
	sc->jumbo_mem_size = size;
	sc->jumbo_mem_ptr = ptr = contigmalloc(size, M_DEVBUF,
		M_NOWAIT, 0, 0xffffffff, PAGE_SIZE, 0);

	if (ptr == NULL) {
		printf("bc%d: no memory for jumbo buffers!\n", sc->bc_unit);
		return(ENOBUFS);
	}

	QQ_InitQueue(&sc->jumbo_free_mem_q.Container, BC_JUMBO_MEM_BLK); 

	for (i = 0; i < BC_JUMBO_MEM_BLK; i++) {
		jmem_ptr = (bc_jmem_t *) ptr;
		jmem_ptr->sc = sc;
		jmem_ptr->bc_inuse = 0;
		QQ_PushHead(&sc->jumbo_free_mem_q.Container, jmem_ptr);
		ptr += BC_JUMBO_FRAMELEN + ETHER_ALIGN + sizeof(bc_jmem_t);
	}

	return(0);
}

static int bc_free_jumbo_mem(sc)
	struct bc_softc		*sc;
{
	if (sc->jumbo_mem_ptr) {
		contigfree(sc->jumbo_mem_ptr, sc->jumbo_mem_size, M_DEVBUF);
		sc->jumbo_mem_ptr = 0;
	}
	return 0;
}

/*
 * Allocate a jumbo buffer.
 */
static void *bc_jalloc(sc)
	struct bc_softc		*sc;
{
	bc_jmem_t		*jmem_ptr;

	jmem_ptr = QQ_PopHead(&sc->jumbo_free_mem_q.Container);

	if (jmem_ptr == NULL) {
		printf("bc%d: no free jumbo buffers\n", sc->bc_unit);
		return(NULL);
	}
	jmem_ptr->bc_inuse = 1;
	return((void *) (jmem_ptr + 1));
}

/*
 * Adjust usage count on a jumbo buffer. In general this doesn't
 * get used much because our jumbo buffers don't get passed around
 * too much, but it's implemented for correctness.
 */
static void bc_jref(buf, size)
	caddr_t			buf;
	u_int			size;
{
	bc_jmem_t		*jmem_ptr;

	jmem_ptr = (bc_jmem_t *) buf;
	jmem_ptr--;

	if (size != BC_JUMBO_FRAMELEN)
		panic("bc_jref: adjusting refcount of buf of wrong size!");

	if (jmem_ptr->bc_inuse == 0)
		panic("bc_jref: buffer already free!");
	else
		jmem_ptr->bc_inuse++;

	return;
}

/*
 * Release a jumbo buffer.
 */
static void bc_jfree(buf, size)
	caddr_t			buf;
	u_int			size;
{
	bc_jmem_t		*jmem_ptr;
	struct bc_softc		*sc;

	jmem_ptr = (bc_jmem_t *) buf;
	jmem_ptr--;
	sc = jmem_ptr->sc;

	if (size != BC_JUMBO_FRAMELEN)
		panic("bc_jfree: freeing buffer of wrong size!");

	if (jmem_ptr->bc_inuse == 0)
		panic("bc_jfree: buffer already free!");
	else {
		jmem_ptr->bc_inuse--;
		if (jmem_ptr->bc_inuse == 0) {
			QQ_PushTail(&sc->jumbo_free_mem_q.Container, jmem_ptr);
		}
	}

	return;
}

#endif /* BC_PRIVATE_JUMBOS */


LM_STATUS
MM_ReadConfig16(PLM_DEVICE_BLOCK pDevice, LM_UINT32 Offset,
	LM_UINT16 *pValue16)
{
	struct bc_softc *sc = (struct bc_softc *) pDevice;

	*pValue16 = pci_read_config(sc->dev, Offset, 2);
	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_ReadConfig32(PLM_DEVICE_BLOCK pDevice, LM_UINT32 Offset,
	LM_UINT32 *pValue32)
{
	struct bc_softc *sc = (struct bc_softc *) pDevice;

	*pValue32 = pci_read_config(sc->dev, Offset, 4);
	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_WriteConfig16(PLM_DEVICE_BLOCK pDevice, LM_UINT32 Offset,
	LM_UINT16 Value16)
{
	struct bc_softc *sc = (struct bc_softc *) pDevice;

	pci_write_config(sc->dev, Offset, Value16, 2);
	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_WriteConfig32(PLM_DEVICE_BLOCK pDevice, LM_UINT32 Offset,
	LM_UINT32 Value32)
{
	struct bc_softc *sc = (struct bc_softc *) pDevice;

	pci_write_config(sc->dev, Offset, Value32, 4);
	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_AllocateSharedMemory(PLM_DEVICE_BLOCK pDevice, LM_UINT32 BlockSize,
	PLM_VOID *pMemoryBlockVirt, PLM_PHYSICAL_ADDRESS pMemoryBlockPhy,
	LM_BOOL Cached)
{
	PLM_VOID pvirt;
	struct bc_softc *sc = (struct bc_softc *) pDevice;

	pvirt = contigmalloc(BlockSize, M_DEVBUF,
		M_NOWAIT, 0, 0xffffffff, PAGE_SIZE, 0);
	if (pvirt == 0)
		return LM_STATUS_FAILURE;
	pMemoryBlockPhy->High = 0;
	pMemoryBlockPhy->Low = vtophys(pvirt);
	*pMemoryBlockVirt = (PLM_VOID) pvirt;
	sc->mem_list[sc->mem_list_num] = pvirt;
	sc->mem_size_list[sc->mem_list_num++] = BlockSize;
	bzero(pvirt, BlockSize);
	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_AllocateMemory(PLM_DEVICE_BLOCK pDevice, LM_UINT32 BlockSize,
	PLM_VOID *pMemoryBlockVirt)
{
	PLM_VOID pvirt;
	struct bc_softc *sc = (struct bc_softc *) pDevice;

	pvirt = malloc(BlockSize, M_DEVBUF, M_NOWAIT);
	if (!pvirt) {
		goto MM_Alloc_error;
	}
	sc->mem_list[sc->mem_list_num] = pvirt;
	sc->mem_size_list[sc->mem_list_num++] = 0x0; /* 0 indicates malloc */
	bzero(pvirt, BlockSize);
	*pMemoryBlockVirt = pvirt;
	return LM_STATUS_SUCCESS;

MM_Alloc_error:
	printf("bc%d: Memory allocation failed - buffer parameters may be set too high!\n", sc->bc_unit);
	return LM_STATUS_FAILURE;
}

LM_STATUS
MM_MapMemBase(PLM_DEVICE_BLOCK pDevice)
{
	struct bc_softc *sc = (struct bc_softc *) pDevice;
	int rid;

	rid = PCIR_MAPS;
	sc->bc_res = bus_alloc_resource(sc->dev, SYS_RES_MEMORY, &rid,
	    0, ~0, 1, RF_ACTIVE);

	if (sc->bc_res == NULL) {
		printf ("bc%d: couldn't map memory\n", sc->bc_unit);
		return LM_STATUS_FAILURE;
	}

	sc->bc_btag = rman_get_bustag(sc->bc_res);
	sc->bc_bhandle = rman_get_bushandle(sc->bc_res);
	sc->bc_vhandle = (vm_offset_t)rman_get_virtual(sc->bc_res);

	pDevice->pMappedMemBase = (PLM_VOID) sc->bc_vhandle;
	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_InitializeUmPackets(PLM_DEVICE_BLOCK pDevice)
{
	int i;
	struct bc_softc *sc = (struct bc_softc *) pDevice;
	PUM_PACKET pUmPacket;
	PLM_PACKET pPacket;

	for (i = 0; i < pDevice->RxPacketDescCnt; i++) {
		pPacket = QQ_PopHead(&pDevice->RxPacketFreeQ.Container);
		pUmPacket = (PUM_PACKET) pPacket;
		if (pPacket == 0) {
			printf ("bc%d: Bad RxPacketFreeQ!\n", sc->bc_unit);
		}
		bc_newbuf(sc, pUmPacket);
		if (pUmPacket->mbuf == 0) {
			QQ_PushTail(&sc->rx_out_of_buf_q.Container, pPacket);
			continue;
		}
		QQ_PushTail(&pDevice->RxPacketFreeQ.Container, pPacket);
	}
	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_GetConfig(PLM_DEVICE_BLOCK pDevice)
{
	struct bc_softc		*sc = (struct bc_softc *) pDevice;
	struct ifnet		*ifp = &sc->arpcom.ac_if;

	pDevice->NoTxPseudoHdrChksum = TRUE;
	pDevice->NoRxPseudoHdrChksum = TRUE;
	if (ifp->if_mtu > 1514)
		pDevice->RxJumboDescCnt = BC_JUMBO_DESC_CNT;
	pDevice->RxMtu = ifp->if_mtu + 14;
	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_StartTxDma(PLM_DEVICE_BLOCK pDevice, PLM_PACKET pPacket)
{
	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_CompleteTxDma(PLM_DEVICE_BLOCK pDevice, PLM_PACKET pPacket)
{
	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_IndicateRxPackets(PLM_DEVICE_BLOCK pDevice)
{
	struct bc_softc *sc = (struct bc_softc *) pDevice;
	PLM_PACKET pPacket;
	PUM_PACKET pUmPacket;
	struct mbuf		*m;
	int size;
	struct ether_header	*eh;
	struct ifnet		*ifp = &sc->arpcom.ac_if;

	while (1) {
		pPacket = (PLM_PACKET)
			QQ_PopHead(&pDevice->RxPacketReceivedQ.Container);
		if (pPacket == 0)
			break;

		pUmPacket = (PUM_PACKET) pPacket;
		if ((pPacket->PacketStatus != LM_STATUS_SUCCESS) ||
			((size = pPacket->PacketSize) > pDevice->RxMtu)) {

			ifp->if_ierrors++;
			QQ_PushTail(&pDevice->RxPacketFreeQ.Container, pPacket);
			continue;
		}
		m = pUmPacket->mbuf;
		if (bc_newbuf(sc, pUmPacket) == ENOBUFS) {
			ifp->if_ierrors++;
			QQ_PushTail(&pDevice->RxPacketFreeQ.Container, pPacket);
			continue;
		}

		m->m_pkthdr.len = m->m_len = pPacket->PacketSize;

		ifp->if_ipackets++;
		if (m->m_len == 0)
			eh = mtod(m->m_next, struct ether_header *);
		else
			eh = mtod(m, struct ether_header *);
		m->m_pkthdr.rcvif = ifp;

		/* Remove header from mbuf and pass it on. */
		m_adj(m, sizeof(struct ether_header));

		if (ifp->if_hwassist) {
			m->m_pkthdr.csum_flags |= CSUM_IP_CHECKED |
			    CSUM_DATA_VALID;
			if ((pPacket->u.Rx.IpChecksum ^ 0xffff) == 0)
				m->m_pkthdr.csum_flags |= CSUM_IP_VALID;
			m->m_pkthdr.csum_data = pPacket->u.Rx.TcpUdpChecksum;
		}
#if NVLAN > 0
		if (pPacket->Flags & RCV_BD_FLAG_VLAN_TAG) {
		/*
		 * If we received a packet with a vlan tag, pass it
		 * to vlan_input() instead of ether_input().
		 */
			vlan_input_tag(eh, m, pPacket->VlanTag);
		}
		else
#endif
		{
			ether_input(ifp, eh, m);
		}

		QQ_PushTail(&pDevice->RxPacketFreeQ.Container, pPacket);
	}
	return LM_STATUS_SUCCESS;
}

#if 0
static int
replenish_rx_buffers(struct bc_softc *sc)
{
	PLM_PACKET pPacket;
	PUM_PACKET pUmPacket;
	PLM_DEVICE_BLOCK pDevice = &sc->lm_dev;
	int queue_rx = 0;

	while ((pUmPacket = (PUM_PACKET)
		QQ_PopHead(&sc->rx_out_of_buf_q.Container)) != 0) {
		pPacket = (PLM_PACKET) pUmPacket;
		if (pUmPacket->mbuf) {
			QQ_PushTail(&pDevice->RxPacketFreeQ.Container, pPacket);
			queue_rx = 1;
			continue;
		}
		if (bc_newbuf(sc, pUmPacket) == ENOBUFS) {
			QQ_PushHead(&sc->rx_out_of_buf_q.Container,
				pPacket);
			break;
		}
	}
	if (queue_rx) {
		LM_QueueRxPackets(pDevice);
	}
	return 0;
}
#endif

LM_STATUS
MM_IndicateTxPackets(PLM_DEVICE_BLOCK pDevice)
{
	struct bc_softc *sc = (struct bc_softc *) pDevice;
	PLM_PACKET pPacket;
	PUM_PACKET pUmPacket;
	struct ifnet		*ifp;

	ifp = &sc->arpcom.ac_if;

	while (1) {
		pPacket = (PLM_PACKET)
			QQ_PopHead(&pDevice->TxPacketXmittedQ.Container);
		if (pPacket == 0)
			break;
		pUmPacket = (PUM_PACKET) pPacket;

		ifp->if_opackets++;
		if (pUmPacket->mbuf) {
			m_freem(pUmPacket->mbuf);
			pUmPacket->mbuf = 0;
		}
		QQ_PushTail(&pDevice->TxPacketFreeQ.Container, pPacket);
	}
	if (ifp->if_flags & IFF_OACTIVE) {
		if (QQ_GetEntryCnt(&pDevice->TxPacketFreeQ.Container) >=
			(pDevice->TxPacketDescCnt >> 1)) {

			ifp->if_flags &= ~IFF_OACTIVE;
		}
	}
	ifp->if_timer = 0;
	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_IndicateStatus(PLM_DEVICE_BLOCK pDevice, LM_STATUS Status)
{
	struct bc_softc *sc = (struct bc_softc *) pDevice;
	LM_FLOW_CONTROL flow_control;
	struct ifnet		*ifp = &sc->arpcom.ac_if;

	if (!(ifp->if_flags & IFF_RUNNING))
		return LM_STATUS_SUCCESS;

	printf("bc%d: link down\n", sc->bc_unit);
	if (Status == LM_STATUS_LINK_DOWN) {
		printf("bc%d: link down\n", sc->bc_unit);
	}
	else if (Status == LM_STATUS_LINK_ACTIVE) {
		printf("bc%d: link up, ", sc->bc_unit);

		if (pDevice->LineSpeed == LM_LINE_SPEED_10MBPS)
			printf("10 Mbps ");
		else if (pDevice->LineSpeed == LM_LINE_SPEED_100MBPS)
			printf("100 Mbps ");
		else if (pDevice->LineSpeed == LM_LINE_SPEED_1000MBPS)
			printf("1000 Mbps ");

		if (pDevice->DuplexMode == LM_DUPLEX_MODE_FULL)
			printf("full duplex");
		else
			printf("half duplex");

		flow_control = pDevice->FlowControl &
			(LM_FLOW_CONTROL_RECEIVE_PAUSE |
			LM_FLOW_CONTROL_TRANSMIT_PAUSE);
		if (flow_control) {
			if (flow_control & LM_FLOW_CONTROL_RECEIVE_PAUSE) {
				printf(", receive ");
				if (flow_control & LM_FLOW_CONTROL_TRANSMIT_PAUSE)
					printf("& transmit ");
			}
			else {
				printf(", transmit ");
			}
			printf("flow control ON");
		}
		printf("\n");
	}
	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_FreeRxBuffer(PLM_DEVICE_BLOCK pDevice, PLM_PACKET pPacket)
{
	PUM_PACKET pUmPacket;

	if (pPacket == 0)
		return LM_STATUS_SUCCESS;
	pUmPacket = (PUM_PACKET) pPacket;
	if (pUmPacket->mbuf)
		m_freem(pUmPacket->mbuf);
	pUmPacket->mbuf = 0;
	return LM_STATUS_SUCCESS;
}

LM_DEVICE_BLOCK *
MM_FindPeerDev(LM_DEVICE_BLOCK *pDevice)
{
	return 0;
}

